package pl.webnet.JaTrac;

class Camera implements XmlSettable
{
	protected Point origin;
	Point target;
	Vec direction;
	Vec up;
	Vec right;
	double angleX;
	double angleY;
	double aspect;
	int recursionDepth = 0;
	Screen screen;
	Point lowerLeft;
	Vec imagePlaneRight, imagePlaneUp;
	RayTree rayTree;

	public Camera()
	{
	}

	public Camera(Point origin, Point target, Vec up, double angle, double aspect)
	{
		this.origin = new Point(origin);
		this.target = new Point(target);
		this.up = new Vec(up);
		up.unit();
		this.aspect = aspect;
		this.angleX = angle;
		this.angleY = angle / aspect;
	}

	public String[] getParamArray()
	{
		final String params[] = {
			"origin", "origin", "",
			"target", "target", "",
			"up", "up", "",
			"angle", "angleX", "",
			"recursion", "recursionDepth", "",
			"ratio", "", "setAspect"
		};

		return params;
	}

	public void addObject(Object o)
	{
		String params[] = {
			"Screen", "attachScreen"
		};

		ParamHelper.addObject(this, params, o);
	}

	public void setAspect(String val)
	{
		double a = ParamHelper.extractDouble(val);

		this.aspect = a;
		angleY = angleX / a;
	}

	public void recalcParams()
	{
		// calculate camera direction vector
		direction = new Vec(target,  origin);

		// compute size of the image plane in world coordinates;
		double imagePlaneWidth = direction.length() * Math.tan(angleX * Math.PI / 180.0 / 2) * 2;
		double imagePlaneHeight = direction.length() * Math.tan(angleY * Math.PI / 180.0 / 2) * 2;
		// compute size of the single pixel in world coordinates;
		double pixelApertureX = imagePlaneWidth / screen.getWidth() ;
		double pixelApertureY = imagePlaneHeight / screen.getHeight() ;

		System.out.println("pixelApertureX: " + pixelApertureX);
		System.out.println("pixelApertureY: " + pixelApertureY);

		direction.unit();

		// compute cross product of direction and up vectors to get a vector pointing right
		right = up.cross(direction);
		right.unit();

		imagePlaneWidth /= 2;
		imagePlaneHeight /= 2;
		
		// compute lower-right corner of image plane in world coordinates
		lowerLeft = new Point(
			target.x + (-imagePlaneWidth * right.x) + (-imagePlaneHeight * up.x) - origin.x,
			target.y + (-imagePlaneWidth * right.y) + (-imagePlaneHeight * up.y) - origin.y,
			target.z + (-imagePlaneWidth * right.z) + (-imagePlaneHeight * up.z) - origin.z);
		
		System.out.println("lowerLeft: " + lowerLeft);

		imagePlaneRight = new Vec(right);
		imagePlaneRight.mul(pixelApertureX);

		imagePlaneUp = new Vec(up);
		imagePlaneUp.mul(pixelApertureY);
		
		System.out.println("imagePlaneRight:" + imagePlaneRight);
		System.out.println("imagePlaneUp:" + imagePlaneUp);
	}
	
	public void attachScreen(Screen scr)
	{
		screen = scr;
		screen.init();
		recalcParams();
	}
	
/**
 * Generate a primary (camera) ray.
 * @param		x point on x axis that ray has to go through (screen space)
 * @param		y point on y axis that ray has to go through (screen space)
 * @return		primary ray originating at camera's position and going through <code>(x, y)</code>
 */

	public Ray generateRay(double x, double y)
	{

		Vec dir = new Vec(
			lowerLeft.x + (x * imagePlaneRight.x) + (y * imagePlaneUp.x),
			lowerLeft.y + (x * imagePlaneRight.y) + (y * imagePlaneUp.y),
			lowerLeft.z + (x * imagePlaneRight.z) + (y * imagePlaneUp.z)
			);


//		System.out.println("raydir(" + x + ", " + y + "): " + dir);
		dir.unit();
		
		return new Ray(origin, dir);
	}

	public void render(Scene scene)
	{
		int width = screen.getWidth();
		int height = screen.getHeight();
		Ray primary;

		rayTree = new RayTree(scene, recursionDepth);
		Spectrum lineCache[][] = new Spectrum[2][width + 1];
		

		for(int i = 0; i < height; i++)
		{
			System.out.println("Processing line: " + i);

			// shoot rays through pixel corners (hence "width + 1")
			for(int j = 0; j < width + 1; j++)
			{
				primary = generateRay((double)j, (double)i);

				lineCache[i % 2][j] = rayTree.trace(0, 1., primary).spectrum;
			}

			// if we have at least two lines computed,
			if(i > 0)
			{
				Spectrum totalSpectrum = null;
				
				for(int j = 0; j < width; j++)
				{
				//	System.out.println("pixel=" + j);

					totalSpectrum = antialiasPixel(0.8, j + 0.5, i - 0.5, 0.5,
														lineCache[(i + 1) % 2][j],
														lineCache[(i + 1) % 2][j + 1],
														lineCache[i % 2][j],
														lineCache[i % 2][j + 1]);

					screen.pixels[i][j] = new Pixel(totalSpectrum);
				}
			}
		}
	}

	protected Spectrum antialiasPixel(double weight, double x, double y, double size, Spectrum sA, Spectrum sB, Spectrum sC, Spectrum sD)
	{
		double threshold = weight;
		Spectrum sAB, sBC, sCD, sDA;

		sAB = sBC = sCD = sDA = null;

		// clone spectral values to avoid modifying the originals
		try {
			sA = (Spectrum)sA.clone();
			sB = (Spectrum)sB.clone();
			sC = (Spectrum)sC.clone();
			sD = (Spectrum)sD.clone();
		} catch (CloneNotSupportedException e) {}

		// no difference can be bigger than this.
		if(weight > 1)
		{
			sA.add(sB);
			sA.add(sC);
			sA.add(sD);

			sA.scale(0.25);	// 1/4
			return sA;
		}


	//	System.out.println("Entering antialiasPixel: weight=" + weight);

		// shoot ray through pixel's center
		Spectrum sE = rayTree.trace(0, 1., generateRay(x, y)).spectrum;
		double halfSize = size / 2;
		

		// check if differences between corners and center are large enough
		// to subdivide further.

		// upper-left
		if(sA.difference(sE) > threshold)
			sA = antialiasPixel(weight * 4, x - halfSize, y - halfSize, halfSize,
				sA,
				sAB = rayTree.trace(0, 1., generateRay(x, y - size)).spectrum,
				sE,
				sDA = rayTree.trace(0, 1., generateRay(x - size, y)).spectrum);
		else
			sA.average(sE);

		// upper-right
		if(sB.difference(sE) > threshold)
			sB = antialiasPixel(weight * 4, x + halfSize, y - halfSize, halfSize,
				sAB != null ? sAB : rayTree.trace(0, 1., generateRay(x, y - size)).spectrum,
				sB,
				sBC = rayTree.trace(0, 1., generateRay(x + size, y)).spectrum,
				sE);
		else
			sB.average(sE);

		// bottom-right
		if(sC.difference(sE) > threshold)
			sC = antialiasPixel(weight * 4, x + halfSize, y + halfSize, halfSize,
				sE,
				sBC != null ? sBC : rayTree.trace(0, 1., generateRay(x + size, y)).spectrum,
				sC,
				sCD = rayTree.trace(0, 1., generateRay(x, y + size)).spectrum);
		else
			sC.average(sE);

		// bottom-left
		if(sD.difference(sE) > threshold)
			sD = antialiasPixel(weight * 4, x - halfSize, y + halfSize, halfSize,
				sDA != null ? sDA : rayTree.trace(0, 1., generateRay(x - size, y)).spectrum,
				sE,
				sCD != null ? sCD : rayTree.trace(0, 1., generateRay(x, y + size)).spectrum,
				sD);
		else
			sD.average(sE);


		// average four corners
		sA.add(sB);
		sA.add(sC);
		sA.add(sD);

		sA.scale(0.25);	// 1/4

		return sA;
	}
}
